/* execute.c -- Executing the scenario */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "locksim.h"

/*--------------------------------------------------------*/

/* Calculates the value of an arithmetic expression */
static int evaluate(struct TreeNode *t, struct Environment *e) {
    if (t->type == NUMBER)
        return t->leaf_value;
    else if (t->type == ID)
        return get_variable_value(e, t->leaf_value);
    else if (t->type == '+')
        return evaluate(t->args[0], e) + evaluate(t->args[1], e);
    else if (t->type == '-')
        return evaluate(t->args[0], e) - evaluate(t->args[1], e);
    else if (t->type == '*')
        return evaluate(t->args[0], e) * evaluate(t->args[1], e);
    else if (t->type == '/')
        return evaluate(t->args[0], e) / evaluate(t->args[1], e);
    else if (t->type == '-')
        return evaluate(t->args[0], e) - evaluate(t->args[1], e);
    else {
        fatal_error("Tried to evaluate unknown node type");
        return -1;
    }
} /* evaluate */

static void execute_setcommand(struct TreeNode *c, struct Environment *e) {
    assert(c->type == '=');
    assert(c->args[0]->type == ID);
    set_variable_value(e, c->args[0]->leaf_value, evaluate(c->args[1], e));
}

void execute_scenario(struct TreeNode *scenario, int randomize, int verbosity) {
    struct TreeNode *initcommands = car(scenario);
    struct TreeNode *transaction_list = cdr(scenario);
    struct Database *database = create_database();
    int nr_transactions = 0;

    while (transaction_list != NULL) {
        assert(transaction_list->type == ';');
        struct TreeNode *this_transaction_tree = car(transaction_list);
        assert(this_transaction_tree->type == TRANSACTION);
        struct Transaction *t = create_transaction(this_transaction_tree->args[0],
                                                   this_transaction_tree->args[1]);

        if (find_transaction_by_name(this_transaction_tree->args[0]->leaf_value) != NULL)
            fatal_error("Duplicate transaction name '%s'",
                        get_symbol_name_unsafe(this_transaction_tree->args[0]->leaf_value));
        add_transaction(t);
        transaction_list = cdr(transaction_list);
        ++nr_transactions;
    }

    if (verbosity >= 1) {
        printf("[Found %d transaction%s to run: ", nr_transactions, nr_transactions == 1 ? "" : "s");
        show_transactions_compact();
        printf("]\n");
    }

    while (initcommands != NULL) {
        struct TreeNode *c = car(initcommands);
        initcommands = cdr(initcommands);
        execute_setcommand(c, database->environment);
    }

    if (verbosity >= 2) {
        printf("The database before transactions:\n");
        show_environment(database->environment);
    }

    struct Transaction *t;
    while ((t = get_random_runnable_transaction()) != NULL) {
        assert(t->state == RUNNABLE);

        /* Duplicate code */
        if (t->rest_of_program == NULL) {
            if (verbosity >= 2)
                printf("[Transaction %s finished]\n", get_symbol_name_unsafe(t->name->leaf_value));
            finish_transaction(t);
        }
        else {
            struct TreeNode *instruction = car(t->rest_of_program);

            if (verbosity >= 3) {
                show_scheduler_state();
                printf("[Executing an instruction in %s: ", get_symbol_name_unsafe(t->name->leaf_value));
                print_compact(instruction);
                printf("]\n");
            }

            if (instruction->type == '=') {
                execute_setcommand(instruction, t->environment);
            }
            else if (instruction->type == READ) {
                int what_variable = instruction->args[0]->leaf_value;
                int value = get_variable_value(database->environment, what_variable);
                set_variable_value(t->environment, what_variable, value);
                if (verbosity >= 1 && find_my_lock(database, what_variable, t) == UNLOCKED) {
                    printf("*** Warning! *** Transaction %s read %s without a lock.\n",
                           get_symbol_name_unsafe(t->name->leaf_value),
                           get_symbol_name_unsafe(what_variable));
                }
            }
            else if (instruction->type == WRITE) {
                int what_variable = instruction->args[0]->leaf_value;
                int value = get_variable_value(t->environment, what_variable);
                set_variable_value(database->environment, what_variable, value);
                if (verbosity >= 1 && find_my_lock(database, what_variable, t) == UNLOCKED) {
                    printf("*** Warning! *** Transaction %s wrote %s without an exclusive lock.\n",
                           get_symbol_name_unsafe(t->name->leaf_value),
                           get_symbol_name_unsafe(what_variable));
                }
            }
            else if (instruction->type == LOCK) {
                int what_variable = instruction->args[0]->leaf_value;
                lock_variable(database, what_variable, t, EXCLUSIVE);
            }
            else if (instruction->type == READLOCK) {
                int what_variable = instruction->args[0]->leaf_value;
                lock_variable(database, what_variable, t, SHARED);
            }
            else if (instruction->type == WRITELOCK) {
                int what_variable = instruction->args[0]->leaf_value;
                lock_variable(database, what_variable, t, EXCLUSIVE);
            }
            else if (instruction->type == UNLOCK) {
                int what_variable = instruction->args[0]->leaf_value;
                lock_variable(database, what_variable, t, UNLOCKED);
            }
            else {
                assert(0); /* This can never happen */
            }

            // print_tree(           t->program , 6);
            // printf("*********************************************************\n");
            // print_tree(           t->rest_of_program , 6);
            // printf("*********************************************************\n");
            // print_tree( instruction , 6);
            // printf("*********************************************************\n");

            if (t->state == BLOCKED) {
                /* We used to be runnable (since we just executed an instruction), but now we got blocked */
                if (verbosity >= 3) {
                    printf("[Transaction %s blocked waiting for %s]\n",
                           get_symbol_name_unsafe(t->name->leaf_value),
                           get_symbol_name_unsafe(instruction->args[0]->leaf_value));
                }
            }
            else {
                /* Not blocked. But maybe finished? */
                t->rest_of_program = cdr(t->rest_of_program);

                /* Duplicate code */
                if (t->rest_of_program == NULL) {
                    if (verbosity >= 2)
                        printf("[Transaction %s finished]\n", get_symbol_name_unsafe(t->name->leaf_value));
                    finish_transaction(t);
                }
            }

            // print_tree(           t->rest_of_program , 6);
            // printf("*********************************************************\n");
        }

    } /* while there are runnable transactions */

    if (verbosity >= 1) {
        printf("\n");
        printf("The scenario has finished.\n");

        if (get_nr_blocked_transactions() > 0)
            printf("*** Deadlock! ***\n");
    
        printf("The database after transactions:\n");
        show_environment_compact(database->environment);

        show_transactions();

        if (get_nr_blocked_transactions() > 0) {
            show_locks(database);
            printf("*** Deadlock! ***\n");
        }

        printf("\n");
    }

    register_result(database->environment, get_nr_blocked_transactions() > 0);
    clear_scheduler();
    destroy_database(database);

} /* execute_scenario */




























